unit KOLjanFXcut;

interface

uses Windows, Kol;

type
  TFilterProc = function(Value: single): single;

  function SplineFilter(Value: single): single;
  function BellFilter(Value: single): single;
  function TriangleFilter(Value: single): single;
  function BoxFilter(Value: single): single;
  function HermiteFilter(Value: single): single;
  function Lanczos3Filter(Value: single): single;
  function MitchellFilter(Value: single): single;

const
    MaxPixelCount = 32768;

procedure Strecth(var Src,Dst: PBitmap; Filter: TFilterProc; FWidth: single);

implementation

type
  TRGBTripleArray = array[0..pred(MaxPixelCount)] of TRGBTriple;
  PRGBTripleArray = ^TRGBTripleArray;
  TFColor = record
    B,G,R: byte;
  end;

function BellFilter(Value: single): single;
begin
 if (Value<0.0) then Value:=-Value;
  if (Value<0.5) then Result:=0.75-Sqr(Value) else
     if (Value<1.5) then
       begin
         Value:=Value-1.5;
         Result:=0.5*Sqr(Value);
       end
     else Result:=0.0;
end;

function BoxFilter(Value: single): single;
begin
  if (Value>-0.5) and (Value<=0.5) then Result:=1.0 else Result:=0.0;
end;

function HermiteFilter(Value: single): single;
begin
  // f(t) = 2|t|^3 - 3|t|^2 + 1, -1 <= t <= 1
  if Value<0.0 then Value:=-Value;
  if Value<1.0 then Result:=(2.0*Value-3.0)*Sqr(Value)+1.0 else Result:=0.0;
end;

function Lanczos3Filter(Value: single): single;
  function SinC(Value: single): single;
  begin
    if Value<>0.0 then
      begin
        Value:=Value*Pi;
        Result:=Sin(Value)/Value;
      end
    else Result:=1.0;
  end;
begin
  if (Value<0.0) then Value:=-Value;
  if (Value<3.0) then Result:=SinC(Value)*SinC(Value/3.0) else Result:=0.0;
end;

function MitchellFilter(Value: single): single;
const
  B=(1.0/3.0);
  C=(1.0/3.0);
var TT: single;
begin
  if (Value<0.0) then Value:=-Value;
  TT:=Sqr(Value);
  if (Value<1.0) then
    begin
      Value:=(((12.0-9.0*B-6.0*C)*(Value*TT))+((-18.0+12.0*B+6.0*C)*TT)+(6.0-2*B));
      Result:=Value/6.0;
    end
  else if (Value<2.0) then
    begin
      Value:=(((-1.0*B-6.0*C)*(Value*TT))+((6.0*B+30.0*C)*TT)+((-12.0*B-48.0*C)*Value)+(8.0*B+24*C));
      Result:=Value/6.0;
    end
  else Result:=0.0;
end;

function SplineFilter(Value: single): single;
var TT: single;
begin
  if (Value<0.0) then Value:=-Value;
  if (Value<1.0) then
    begin
      TT:=Sqr(Value);
      Result:=0.5*TT*Value-TT+2.0/3.0;
    end
  else if (Value<2.0) then
    begin
      Value:=2.0-Value;
      Result:=1.0/6.0*Sqr(Value)*Value;
    end
  else Result:=0.0;
end;

function TriangleFilter(Value: single): single;
begin
  if (Value<0.0) then Value:=-Value;
  if (Value<1.0) then Result:=1.0-Value else Result:=0.0;
end;



type

  TContributor = record
    pixel: integer;		// Source pixel
    weight: single;		// Pixel weight
  end;

  TContributorList = array[0..0] of TContributor;
  PContributorList = ^TContributorList;

  TCList = record
    n: integer;
    p: PContributorList;
  end;

  TCListList = array[0..0] of TCList;
  PCListList = ^TCListList;

  TRGB = packed record
    r,g,b :single;
  end;

  TColorRGB = packed record
    r,g,b : byte;
  end;
  PColorRGB = ^TColorRGB;

  TRGBList = packed array[0..0] of TColorRGB;
  PRGBList = ^TRGBList;        

procedure Strecth(var Src,Dst: PBitmap; Filter: TFilterProc; FWidth: single);
var
  xscale, yscale	: single;		// Zoom scale factors
  i, j, k		: integer;		// Loop variables
  center		: single;		// Filter calculation variables
  width, fscale, weight	: single;		// Filter calculation variables
  left, right		: integer;		// Filter calculation variables
  n			: integer;		// Pixel number
  Work			: PBitmap;
  contrib		: PCListList;
  rgb			: TRGB;
  color			: TColorRGB;

  SourceLine		,
  DestLine		: PRGBList;
  SourcePixel		,
  DestPixel		: PColorRGB;
  Delta			,
  DestDelta		: integer;

  SrcWidth		,
  SrcHeight		,
  DstWidth		,
  DstHeight		: integer;


function Floor(X: Extended): Integer;
begin
  Result := Integer(Trunc(X));
  if Frac(X) < 0 then
    Dec(Result);
end;

function Ceil(X: Extended): Integer;
begin
  Result := Integer(Trunc(X));
  if Frac(X) > 0 then
    Inc(Result);
end;

begin
  DstWidth:=Dst.Width;
  DstHeight:=Dst.Height;
  SrcWidth:=Src.Width;
  SrcHeight:=Src.Height;
  try
    Work:=NewBitmap(DstWidth,SrcHeight);

    if SrcWidth=1 then xscale:=DstWidth/SrcWidth else xscale:=(DstWidth-1)/(SrcWidth-1);
    if SrcHeight=1 then yscale:=DstHeight/SrcHeight else yscale:=(DstHeight-1)/(SrcHeight-1);


    Src.PixelFormat:=pf24bit;
    Dst.PixelFormat:=Src.PixelFormat;
    Work.PixelFormat:=Src.PixelFormat;

    GetMem(contrib, DstWidth* sizeof(TCList));

    if (xscale < 1.0) then
    begin
      width:=fwidth / xscale;
      fscale:=1.0 / xscale;
      for i:=0 to DstWidth-1 do
      begin
        contrib^[i].n:=0;
        GetMem(contrib^[i].p, trunc(width * 2.0+1) * sizeof(TContributor));
        center:=i / xscale;
        // Original code:
        // left:=ceil(center - width);
        // right:=floor(center+width);
        left:=floor(center - width);
        right:=ceil(center+width);
        for j:=left to right do
        begin
          weight:=filter((center-j)/fscale)/fscale;
          if weight=0.0 then continue;
          if j<0 then n:=-j else
            if j>=SrcWidth then n:=SrcWidth-j+SrcWidth-1 else n:=j;
          k:=contrib^[i].n;
          contrib^[i].n:=contrib^[i].n+1;
          contrib^[i].p^[k].pixel:=n;
          contrib^[i].p^[k].weight:=weight;
        end;
      end;
    end else
    begin
      for i:=0 to DstWidth-1 do
      begin
        contrib^[i].n:=0;
        GetMem(contrib^[i].p,trunc(fwidth*2.0+1)*sizeof(TContributor));
        center:=i/xscale;
        // Original code:
        // left:=ceil(center - fwidth);
        // right:=floor(center+fwidth);
        left:=floor(center-fwidth);
        right:=ceil(center+fwidth);
        for j:=left to right do
        begin
          weight:=filter(center-j);
          if weight=0.0 then continue;
          if j<0 then n:=-j else
            if j>=SrcWidth then n:=SrcWidth-j+SrcWidth-1 else n:=j;
          k:=contrib^[i].n;
          contrib^[i].n:=contrib^[i].n+1;
          contrib^[i].p^[k].pixel:=n;
          contrib^[i].p^[k].weight:=weight;
        end;
      end;
    end;

    for k:=0 to SrcHeight-1 do
    begin

      SourceLine:=Src.ScanLine[k];
      DestPixel:=Work.ScanLine[k];

      for i:=0 to DstWidth-1 do
      begin
        rgb.r:=0.0;
        rgb.g:=0.0;
        rgb.b:=0.0;
        for j:=0 to contrib^[i].n-1 do
        begin

          color:=SourceLine^[contrib^[i].p^[j].pixel];

          weight:=contrib^[i].p^[j].weight;
          if weight=0.0 then continue;
          rgb.r:=rgb.r+color.r*weight;
          rgb.g:=rgb.g+color.g*weight;
          rgb.b:=rgb.b+color.b*weight;
        end;
        if rgb.r>255.0 then color.r:=255 else
          if rgb.r<0.0 then color.r:=0 else color.r:=round(rgb.r);
        if rgb.g>255.0 then color.g:=255 else
          if rgb.g<0.0 then color.g:=0 else color.g:=round(rgb.g);
        if rgb.b>255.0 then color.b:=255 else
          if rgb.b<0.0 then color.b:=0 else color.b:=round(rgb.b);

        DestPixel^:=color;
        inc(DestPixel);

      end;
    end;
    // Free the memory allocated for horizontal filter weights
    for i:=0 to DstWidth-1 do FreeMem(contrib^[i].p);
    FreeMem(contrib);
    // -----------------------------------------------
    // Pre-calculate filter contributions for a column
    // -----------------------------------------------
    GetMem(contrib,DstHeight*sizeof(TCList));
    // Vertical sub-sampling
    // Scales from bigger to smaller height
    if yscale<1.0 then
    begin
      width:=fwidth/yscale;
      fscale:=1.0/yscale;
      for i:=0 to DstHeight-1 do
      begin
        contrib^[i].n:=0;
        GetMem(contrib^[i].p,trunc(width*2.0+1)*sizeof(TContributor));
        center:=i/yscale;
        // Original code:
        // left:=ceil(center-width);
        // right:=floor(center+width);
        left:=floor(center-width);
        right:=ceil(center+width);
        for j:=left to right do
        begin
          weight:=filter((center-j)/fscale)/fscale;
          if weight=0.0 then continue;
          if j<0 then n:=-j else
            if j>=SrcHeight then n:=SrcHeight-j+SrcHeight-1 else n:=j;
          k:=contrib^[i].n;
          contrib^[i].n:=contrib^[i].n+1;
          contrib^[i].p^[k].pixel:=n;
          contrib^[i].p^[k].weight:=weight;
        end;
      end
    end else
    // Vertical super-sampling
    // Scales from smaller to bigger height
    begin
      for i:=0 to DstHeight-1 do
      begin
        contrib^[i].n:=0;
        GetMem(contrib^[i].p,trunc(fwidth*2.0+1)*sizeof(TContributor));
        center:=i / yscale;
        // Original code:
        // left:=ceil(center-fwidth);
        // right:=floor(center+fwidth);
        left:=floor(center-fwidth);
        right:=ceil(center+fwidth);
        for j:=left to right do
        begin
          weight:=filter(center-j);
          if weight=0.0 then continue;
          if j<0 then n:=-j else
            if j>=SrcHeight then n:=SrcHeight-j+SrcHeight-1 else n:=j;
          k:=contrib^[i].n;
          contrib^[i].n:=contrib^[i].n+1;
          contrib^[i].p^[k].pixel:=n;
          contrib^[i].p^[k].weight:=weight;
        end;
      end;
    end;
    // --------------------------------------------------
    // Apply filter to sample vertically from Work to Dst
    // --------------------------------------------------

    SourceLine:=Work.ScanLine[0];
    Delta:=integer(Work.ScanLine[1])-integer(SourceLine);
    DestLine:=Dst.ScanLine[0];
    DestDelta:=integer(Dst.ScanLine[1])-integer(DestLine);

    for k:=0 to DstWidth-1 do
    begin

      DestPixel:=pointer(DestLine);

      for i:=0 to DstHeight-1 do
      begin
        rgb.r:=0;
        rgb.g:=0;
        rgb.b:=0;
        // weight:=0.0;
        for j:=0 to contrib^[i].n-1 do
        begin

          color:=PColorRGB(integer(SourceLine)+contrib^[i].p^[j].pixel*Delta)^;

          weight:=contrib^[i].p^[j].weight;
          if (weight = 0.0) then
            continue;
          rgb.r:=rgb.r+color.r * weight;
          rgb.g:=rgb.g+color.g * weight;
          rgb.b:=rgb.b+color.b * weight;
        end;
        if rgb.r>255.0 then color.r:=255 else
          if rgb.r<0.0 then color.r:=0 else color.r:=round(rgb.r);
        if rgb.g>255.0 then color.g:=255 else
          if rgb.g<0.0 then color.g:=0 else color.g:=round(rgb.g);
        if rgb.b>255.0 then color.b:=255 else
          if rgb.b<0.0 then color.b:=0 else color.b:=round(rgb.b);

        DestPixel^:=color;
        inc(integer(DestPixel),DestDelta);

      end;

      Inc(SourceLine);
      Inc(DestLine);

    end;
    // Free the memory allocated for vertical filter weights
    for i:=0 to DstHeight-1 do FreeMem(contrib^[i].p);
    FreeMem(contrib);
  finally
    Work.Free;
  end;
end;

end.


